@@ -1,13 +1,23 @@
-import os
+from __future__ import annotations
+
+from typing import Optional
+
+from heudiconv.utils import SeqInfo
 
 
-def create_key(template, outtype=('nii.gz',), annotation_classes=None):
+def create_key(
+    template: Optional[str],
+    outtype: tuple[str, ...] = ("nii.gz",),
+    annotation_classes: None = None,
+) -> tuple[str, tuple[str, ...], None]:
     if template is None or not template:
-        raise ValueError('Template must be a valid format string')
-    return template, outtype, annotation_classes
+        raise ValueError("Template must be a valid format string")
+    return (template, outtype, annotation_classes)
 
 
-def infotodict(seqinfo):
+def infotodict(
+    seqinfo: list[SeqInfo],
+) -> dict[tuple[str, tuple[str, ...], None], list[str]]:
     """Heuristic evaluator for determining which runs belong where
 
     allowed template fields - follow python string module:
@@ -18,9 +28,8 @@ def infotodict(seqinfo):
     subindex: sub index within group
     """
 
-    data = create_key('run{item:03d}')
-    info = {data: []}
-    last_run = len(seqinfo)
+    data = create_key("run{item:03d}")
+    info: dict[tuple[str, tuple[str, ...], None], list[str]] = {data: []}
 
     for s in seqinfo:
         """
