@@ -1,93 +1,112 @@
 """Utility objects and functions"""
-import hashlib
-import os
-import tempfile
-import json
-import re
-import sys
-import shutil
+from __future__ import annotations
+
+from collections.abc import Callable
+from collections.abc import Mapping as MappingABC
 import copy
-import stat
+from datetime import datetime
+from glob import glob
+import hashlib
+import json
+from json.decoder import JSONDecodeError
+import logging
+import os
 import os.path as op
 from pathlib import Path
-from collections import namedtuple
-from glob import glob
+import re
+import shutil
+import stat
 from subprocess import check_output
-from datetime import datetime
+import sys
+import tempfile
 from time import sleep
-
-from nipype.utils.filemanip import which
-
-import logging
-lgr = logging.getLogger(__name__)
-
-from json.decoder import JSONDecodeError
-
-
-seqinfo_fields = [
-    'total_files_till_now',  # 0
-    'example_dcm_file',      # 1
-    'series_id',             # 2
-    'dcm_dir_name',          # 3
-    'series_files',          # 4
-    'unspecified',           # 5
-    'dim1', 'dim2', 'dim3', 'dim4', # 6, 7, 8, 9
-    'TR', 'TE',              # 10, 11
-    'protocol_name',         # 12
-    'is_motion_corrected',   # 13
-    'is_derived',            # 14
-    'patient_id',            # 15
-    'study_description',     # 16
-    'referring_physician_name', # 17
-    'series_description',    # 18
-    'sequence_name',         # 19
-    'image_type',            # 20
-    'accession_number',      # 21
-    'patient_age',           # 22
-    'patient_sex',           # 23
-    'date',                  # 24
-    'series_uid',            # 25
-    'time',                  # 26
-]
-
-SeqInfo = namedtuple('SeqInfo', seqinfo_fields)
-
-StudySessionInfo = namedtuple(
-    'StudySessionInfo',
-    [
-        'locator',  # possible prefix identifying the study, e.g.
-                    # PI/dataset or just a dataset or empty (default)
-                    # Note that ATM there should be no multiple DICOMs with the
-                    # same StudyInstanceUID which would collide, i.e point to
-                    # the same subject/session. So 'locator' is pretty much an
-                    # assignment from StudyInstanceUID into some place within
-                    # hierarchy
-        'session',  # could be None
-        'subject',  # should be some ID defined either in cmdline or deduced
-    ]
+from types import ModuleType
+from typing import (
+    Any,
+    AnyStr,
+    Mapping,
+    NamedTuple,
+    Optional,
+    Sequence,
+    Tuple,
+    TypeVar,
+    Union,
+    cast,
+    overload,
 )
 
+lgr = logging.getLogger(__name__)
 
-class TempDirs(object):
+T = TypeVar("T")
+V = TypeVar("V")
+
+
+class SeqInfo(NamedTuple):
+    total_files_till_now: int  # 0
+    example_dcm_file: str  # 1
+    series_id: str  # 2
+    dcm_dir_name: str  # 3
+    series_files: int  # 4
+    unspecified: str  # 5
+    dim1: int  # 6
+    dim2: int  # 7
+    dim3: int  # 8
+    dim4: int  # 9
+    TR: float  # 10
+    TE: float  # 11
+    protocol_name: str  # 12
+    is_motion_corrected: bool  # 13
+    is_derived: bool  # 14
+    patient_id: Optional[str]  # 15
+    study_description: str  # 16
+    referring_physician_name: str  # 17
+    series_description: str  # 18
+    sequence_name: str  # 19
+    image_type: tuple[str, ...]  # 20
+    accession_number: str  # 21
+    patient_age: Optional[str]  # 22
+    patient_sex: Optional[str]  # 23
+    date: Optional[str]  # 24
+    series_uid: Optional[str]  # 25
+    time: Optional[str]  # 26
+
+
+class StudySessionInfo(NamedTuple):
+    # possible prefix identifying the study, e.g.  PI/dataset or just a dataset
+    # or empty (default) Note that ATM there should be no multiple DICOMs with
+    # the same StudyInstanceUID which would collide, i.e point to the same
+    # subject/session. So 'locator' is pretty much an assignment from
+    # StudyInstanceUID into some place within hierarchy
+    locator: Optional[str]
+
+    session: Optional[str]
+
+    # should be some ID defined either in cmdline or deduced
+    subject: Optional[str]
+
+
+class TempDirs:
     """A helper to centralize handling and cleanup of dirs"""
 
-    def __init__(self):
-        self.dirs = []
-        self.exists = op.exists
-        self.lgr = logging.getLogger('tempdirs')
+    def __init__(self) -> None:
+        self.dirs: list[str] = []
+        self.lgr: logging.Logger = logging.getLogger("tempdirs")
 
-    def __call__(self, prefix=None):
+    def __call__(self, prefix: Optional[str] = None) -> str:
         tmpdir = tempfile.mkdtemp(prefix=prefix)
         self.dirs.append(tmpdir)
         return tmpdir
 
-    def __del__(self):
+    def __del__(self) -> None:
         try:
             self.cleanup()
         except AttributeError:
             pass
 
-    def cleanup(self):
+    def exists(self, path: str) -> bool:
+        return op.exists(path)
+
+    def cleanup(self) -> None:
         self.lgr.debug("Removing %d temporary directories", len(self.dirs))
         for t in self.dirs[:]:
             self.lgr.debug("Removing %s", t)
@@ -95,22 +114,25 @@ class TempDirs(object):
                 self.rmtree(t)
         self.dirs = []
 
-    def rmtree(self, tmpdir):
+    def rmtree(self, tmpdir: str) -> None:
         if self.exists(tmpdir):
             shutil.rmtree(tmpdir)
         if tmpdir in self.dirs:
             self.dirs.remove(tmpdir)
 
 
-def docstring_parameter(*sub):
-    """ Borrowed from https://stackoverflow.com/a/10308363/6145776 """
-    def dec(obj):
+def docstring_parameter(*sub: str) -> Callable[[T], T]:
+    """Borrowed from https://stackoverflow.com/a/10308363/6145776"""
+
+    def dec(obj: T) -> T:
+        assert obj.__doc__ is not None
         obj.__doc__ = obj.__doc__.format(*sub)
         return obj
+
     return dec
 
 
-def anonymize_sid(sid, anon_sid_cmd):
+def anonymize_sid(sid: AnyStr, anon_sid_cmd: str) -> AnyStr:
     """
     Raises
     ------
@@ -118,7 +140,7 @@ def anonymize_sid(sid, anon_sid_cmd):
       if script returned an empty string (after whitespace stripping),
       or output with multiple words/lines.
     """
-    cmd = [anon_sid_cmd, sid]
+    cmd: list[str | bytes] = [anon_sid_cmd, sid]
     shell_return = check_output(cmd)
 
     if isinstance(shell_return, bytes) and isinstance(sid, str):
@@ -136,7 +158,7 @@ def anonymize_sid(sid, anon_sid_cmd):
     return anon_sid
 
 
-def create_file_if_missing(filename, content):
+def create_file_if_missing(filename: str, content: str) -> bool:
     """Create file if missing, so we do not
     override any possibly introduced changes"""
     if op.lexists(filename):
@@ -144,29 +166,30 @@ def create_file_if_missing(filename, content):
     dirname = op.dirname(filename)
     if not op.exists(dirname):
         os.makedirs(dirname)
-    with open(filename, 'w') as f:
+    with open(filename, "w") as f:
         f.write(content)
     return True
 
 
-def read_config(infile):
-    with open(infile, 'rt') as fp:
-        info = eval(fp.read())
+def read_config(infile: str) -> dict:
+    with open(infile, "r") as fp:
+        info = cast(dict, eval(fp.read()))
     return info
 
 
-def write_config(outfile, info):
+def write_config(outfile: str, info: dict) -> None:
     from pprint import PrettyPrinter
-    with open(outfile, 'wt') as fp:
+
+    with open(outfile, "w") as fp:
         fp.writelines(PrettyPrinter().pformat(info))
 
 
-def load_json(filename, retry=0):
+def load_json(filename: str | Path, retry: int = 0) -> Any:
     """Load data from a json file
 
     Parameters
     ----------
-    filename : str
+    filename : str or Path
         Filename to load data from.
     retry: int, optional
         Number of times to retry opening/loading the file in case of
@@ -183,7 +206,7 @@ def load_json(filename, retry=0):
     for i in range(retry + 1):  # >= 10 sec wait
         try:
             try:
-                with open(filename, 'r') as fp:
+                with open(filename, "r") as fp:
                     data = json.load(fp)
                     break
             except JSONDecodeError:
@@ -197,20 +220,26 @@ def load_json(filename, retry=0):
             continue
 
     return data
-    
 
-def assure_no_file_exists(path):
+
+def assure_no_file_exists(path: str | Path) -> None:
     """Check if file or symlink (git-annex?) exists, and if so -- remove"""
     if os.path.lexists(path):
         os.unlink(path)
 
 
-def save_json(filename, data, indent=2, sort_keys=True, pretty=False):
+def save_json(
+    filename: str | Path,
+    data: Any,
+    indent: int = 2,
+    sort_keys: bool = True,
+    pretty: bool = False,
+) -> None:
     """Save data to a json file
 
     Parameters
     ----------
-    filename : str
+    filename : str or Path
         Filename to save data in.
     data : dict
         Dictionary to save in json file.
@@ -220,34 +249,31 @@ def save_json(filename, data, indent=2, sort_keys=True, pretty=False):
 
     """
     assure_no_file_exists(filename)
-    dumps_kw = dict(sort_keys=sort_keys, indent=indent)
-    j = None
+    j: Optional[str] = None
     if pretty:
         try:
-            j = json_dumps_pretty(data, **dumps_kw)
+            j = json_dumps_pretty(data, sort_keys=sort_keys, indent=indent)
         except AssertionError as exc:
             pretty = False
             lgr.warning(
                 "Prettyfication of .json failed (%s).  "
                 "Original .json will be kept as is.  Please share (if you "
                 "could) "
-                "that file (%s) with HeuDiConv developers"
-                % (str(exc), filename)
+                "that file (%s) with HeuDiConv developers" % (str(exc), filename)
             )
     if not pretty:
-        j = json_dumps(data, **dumps_kw)
+        j = json_dumps(data, sort_keys=sort_keys, indent=indent)
     assert j is not None  # one way or another it should have been set to a str
-    with open(filename, 'w') as fp:
+    with open(filename, "w") as fp:
         fp.write(j)
 
 
-def json_dumps(json_obj, indent=2, sort_keys=True):
-    """Unified (default indent and sort_keys) invocation of json.dumps
-    """
+def json_dumps(json_obj: Any, indent: int = 2, sort_keys: bool = True) -> str:
+    """Unified (default indent and sort_keys) invocation of json.dumps"""
     return json.dumps(json_obj, indent=indent, sort_keys=sort_keys)
 
 
-def json_dumps_pretty(j, indent=2, sort_keys=True):
+def json_dumps_pretty(j: Any, indent: int = 2, sort_keys: bool = True) -> str:
     """Given a json structure, pretty print it by colliding numeric arrays
     into a line.
 
@@ -256,37 +282,45 @@ def json_dumps_pretty(j, indent=2, sort_keys=True):
     js = json_dumps(j, indent=indent, sort_keys=sort_keys)
     # trim away \n and spaces between entries of numbers
     js_ = re.sub(
-        '[\n ]+("?[-+.0-9e]+"?,?) *\n(?= *"?[-+.0-9e]+"?)', r' \1',
-        js, flags=re.MULTILINE)
+        '[\n ]+("?[-+.0-9e]+"?,?) *\n(?= *"?[-+.0-9e]+"?)',
+        r" \1",
+        js,
+        flags=re.MULTILINE,
+    )
     # uniform no spaces before ]
     js_ = re.sub(r" *\]", "]", js_)
     # uniform spacing before numbers
     # But that thing could screw up dates within strings which would have 2 spaces
     # in a date like Mar  3 2017, so we do negative lookahead to avoid changing
     # in those cases
-    #import pdb; pdb.set_trace()
+    # import pdb; pdb.set_trace()
     js_ = re.sub(
-        r'(?<!\w{3})'    # negative lookbehind for the month
+        r"(?<!\w{3})"  # negative lookbehind for the month
         r'  *("?[-+.0-9e]+"?)'
-        r'(?! [123]\d{3})'  # negative lookahead for a year
-        r'(?P<space> ?)[ \n]*',
-        r' \1\g<space>', js_)
+        r"(?! [123]\d{3})"  # negative lookahead for a year
+        r"(?P<space> ?)[ \n]*",
+        r" \1\g<space>",
+        js_,
+    )
     # no spaces after [
-    js_ = re.sub(r'\[ ', '[', js_)
+    js_ = re.sub(r"\[ ", "[", js_)
     # the load from the original dump and reload from tuned up
     # version should result in identical values since no value
     # must be changed, just formatting.
     j_just_reloaded = json.loads(js)
     j_tuned = json.loads(js_)
 
-    assert j_just_reloaded == j_tuned, \
-       "Values differed when they should have not. "\
-       "Report to the heudiconv developers"
+    assert j_just_reloaded == j_tuned, (
+        "Values differed when they should have not. "
+        "Report to the heudiconv developers"
+    )
 
     return js_
 
 
-def update_json(json_file, new_data, pretty=False):
+def update_json(
+    json_file: str | Path, new_data: dict[str, Any], pretty: bool = False
+) -> None:
     """
     Adds a given field (and its value) to a json file
 
@@ -314,16 +348,15 @@ def update_json(json_file, new_data, pretty=False):
     save_json(json_file, data, pretty=pretty)
 
 
-def treat_infofile(filename):
-    """Tune up generated .json file (slim down, pretty-print for humans).
-    """
+def treat_infofile(filename: str) -> None:
+    """Tune up generated .json file (slim down, pretty-print for humans)."""
     j = load_json(filename)
     j_slim = slim_down_info(j)
     save_json(filename, j_slim, sort_keys=True, pretty=True)
     set_readonly(filename)
 
 
-def slim_down_info(j):
+def slim_down_info(j: dict[str, Any]) -> dict[str, Any]:
     """Given an aggregated info structure, removes excessive details
 
     Such as CSA fields, and SourceImageSequence which on Siemens files could be
@@ -331,89 +364,87 @@ def slim_down_info(j):
     If needed, could be recovered from stored DICOMs
     """
     j = copy.deepcopy(j)  # we will do in-place modification on a copy
-    dicts = []
+    dicts: list[dict[str, Any]] = []
     # poor man programming for now
-    if 'const' in j.get('global', {}):
-        dicts.append(j['global']['const'])
-    if 'samples' in j.get('time', {}):
-        dicts.append(j['time']['samples'])
+    if "const" in j.get("global", {}):
+        dicts.append(j["global"]["const"])
+    if "samples" in j.get("time", {}):
+        dicts.append(j["time"]["samples"])
     for d in dicts:
         for k in list(d.keys()):
-            if k.startswith('Csa') or k.lower() in {'sourceimagesequence'}:
+            if k.startswith("Csa") or k.lower() in {"sourceimagesequence"}:
                 del d[k]
     return j
 
 
-def get_known_heuristic_names():
+def get_known_heuristic_names() -> list[str]:
     """Return a list of heuristic names present under heudiconv/heuristics"""
     import heudiconv.heuristics
+
     candidates = {
         op.splitext(op.basename(x))[0]
         for hp in heudiconv.heuristics.__path__
-        for x in glob(op.join(hp, '*.py')) + glob(op.join(hp, '*.py[co]'))
+        for x in glob(op.join(hp, "*.py")) + glob(op.join(hp, "*.py[co]"))
     }
     return sorted(
-        filter(
-            lambda c: not (c.startswith('test_') or c.startswith('_')),
-            candidates
-        )
+        filter(lambda c: not (c.startswith("test_") or c.startswith("_")), candidates)
     )
 
 
-def load_heuristic(heuristic):
-    """Load heuristic from the file, return the module
-    """
+def load_heuristic(heuristic: str) -> ModuleType:
+    """Load heuristic from the file, return the module"""
     if os.path.sep in heuristic or os.path.lexists(heuristic):
         heuristic_file = op.realpath(heuristic)
         path, fname = op.split(heuristic_file)
         try:
             old_syspath = sys.path[:]
             sys.path.insert(0, path)
-            mod = __import__(fname.split('.')[0])
-            mod.filename = heuristic_file
+            mod = __import__(fname.split(".")[0])
+            mod.filename = heuristic_file  # type: ignore[attr-defined]
         finally:
             sys.path = old_syspath
     else:
         from importlib import import_module
+
         try:
-            mod = import_module('heudiconv.heuristics.%s' % heuristic)
-            mod.filename = mod.__file__.rstrip('co')  # remove c or o from pyc/pyo
+            mod = import_module("heudiconv.heuristics.%s" % heuristic)
+            assert mod.__file__ is not None
+            # remove c or o from pyc/pyo
+            mod.filename = mod.__file__.rstrip("co")  # type: ignore[attr-defined]
         except Exception as exc:
-            raise ImportError(
-                "Failed to import heuristic %s: %s"
-                % (heuristic, exc)
-            )
+            raise ImportError("Failed to import heuristic %s: %s" % (heuristic, exc))
     return mod
 
 
-def get_heuristic_description(name, full=False):
+def get_heuristic_description(name: str, full: bool = False) -> str:
     try:
         mod = load_heuristic(name)
-        desc = (getattr(mod, '__doc__', '') or '').strip()
+        desc = (getattr(mod, "__doc__", "") or "").strip()
         return desc.split(os.linesep)[0] if not full else desc
     except Exception as exc:
         return "Failed to load: %s" % exc
 
 
-def get_known_heuristics_with_descriptions():
+def get_known_heuristics_with_descriptions() -> dict[str, str]:
     from collections import OrderedDict
+
     heuristics = OrderedDict()
     for name in get_known_heuristic_names():
         heuristics[name] = get_heuristic_description(name, full=False)
     return heuristics
 
 
-def safe_copyfile(src, dest, overwrite=False):
+def safe_copyfile(src: str, dest: str, overwrite: bool = False) -> None:
     """Copy file but blow if destination name already exists"""
-    return _safe_op_file(src, dest, "copyfile", overwrite=overwrite)
+    _safe_op_file(src, dest, "copyfile", overwrite=overwrite)
 
 
-def safe_movefile(src, dest, overwrite=False):
+def safe_movefile(src: str, dest: str, overwrite: bool = False) -> None:
     """Move file but blow if destination name already exists"""
-    return _safe_op_file(src, dest, "move", overwrite=overwrite)
+    _safe_op_file(src, dest, "move", overwrite=overwrite)
 
 
-def _safe_op_file(src, dest, operation, overwrite=False):
+def _safe_op_file(src: str, dest: str, operation: str, overwrite: bool = False) -> None:
     """Copy or move file but blow if destination name already exists
 
     Parameters
@@ -436,11 +467,12 @@ def _safe_op_file(src, dest, operation, overwrite=False):
 
 
 # Globals to check filewriting permissions
-ALL_CAN_WRITE = (stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
-ALL_CAN_READ = (stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
+ALL_CAN_WRITE = stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
+ALL_CAN_READ = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
 assert ALL_CAN_READ >> 1 == ALL_CAN_WRITE  # Assumption in the code
 
-def set_readonly(path, read_only=True):
+
+def set_readonly(path: str, read_only: bool = True) -> int:
     """Make file read only or writeable while preserving "access levels"
 
     So if file was not readable by others, it should remain not readable by
@@ -471,41 +503,43 @@ def set_readonly(path, read_only=True):
     return new_perms
 
 
-def is_readonly(path):
-    """Return True if it is a fully read-only file (dereferences the symlink)
-    """
+def is_readonly(path: str) -> bool:
+    """Return True if it is a fully read-only file (dereferences the symlink)"""
     # get current permissions
     perms = stat.S_IMODE(os.lstat(os.path.realpath(path)).st_mode)
     # should be true if anyone is allowed to write
     return not bool(perms & ALL_CAN_WRITE)
 
 
-def clear_temp_dicoms(item_dicoms):
+def clear_temp_dicoms(item_dicoms: list[str]) -> None:
     """Ensures DICOM temporary directories are safely cleared"""
     try:
         tmp = Path(op.commonprefix(item_dicoms)).parents[1]
     except IndexError:
         return
-    if (str(tmp.parent) == tempfile.gettempdir()
-        and str(tmp.stem).startswith('heudiconvDCM')
-        and op.exists(str(tmp))):
+    if (
+        str(tmp.parent) == tempfile.gettempdir()
+        and str(tmp.stem).startswith("heudiconvDCM")
+        and op.exists(tmp)
+    ):
         # clean up directory holding dicoms
-        shutil.rmtree(str(tmp))
+        shutil.rmtree(tmp)
 
 
-def file_md5sum(filename):
-    with open(filename, 'rb') as f:
+def file_md5sum(filename: str) -> str:
+    with open(filename, "rb") as f:
         return hashlib.md5(f.read()).hexdigest()
 
 
 # Borrowed from DataLad (MIT license), with "archives" functionality commented
 # out
-class File(object):
+class File:
     """Helper for a file entry in the create_tree/@with_tree
 
     It allows to define additional settings for entries
     """
-    def __init__(self, name, executable=False):
+
+    def __init__(self, name: str, executable: bool = False) -> None:
         """
 
         Parameters
@@ -518,11 +552,19 @@ class File(object):
         self.name = name
         self.executable = executable
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.name
 
 
-def create_tree(path, tree, archives_leading_dir=True):
+TreeSpec = Union[
+    Sequence[Tuple[Union[str, File], "Load"]],
+    Mapping[str, "Load"],
+]
+
+Load = Union[str, "TreeSpec", dict]
+
+
+def create_tree(path: str, tree: TreeSpec, archives_leading_dir: bool = True) -> None:
     """Given a list of tuples (name, load) or a dict create such a tree
 
     if load is a tuple or a dict itself -- that would create either a subtree
@@ -533,8 +575,8 @@ def create_tree(path, tree, archives_leading_dir=True):
     if not op.exists(path):
         os.makedirs(path)
 
-    if isinstance(tree, dict):
-        tree = tree.items()
+    if isinstance(tree, MappingABC):
+        tree = list(tree.items())
 
     for file_, load in tree:
         if isinstance(file_, File):
@@ -544,7 +586,7 @@ def create_tree(path, tree, archives_leading_dir=True):
             executable = False
             name = file_
         full_name = op.join(path, name)
-        if name.endswith('.json') and isinstance(load, dict):
+        if name.endswith(".json") and isinstance(load, dict):
             # (For a json file, we expect the content to be a dictionary, so
             #  don't continue creating a tree, but just write dict to file)
             save_json(full_name, load)
@@ -554,13 +596,28 @@ def create_tree(path, tree, archives_leading_dir=True):
             # else:
             create_tree(full_name, load, archives_leading_dir=archives_leading_dir)
         else:
-            with open(full_name, 'w') as f:
+            assert isinstance(load, str)
+            with open(full_name, "w") as f:
                 f.write(load)
         if executable:
             os.chmod(full_name, os.stat(full_name).st_mode | stat.S_IEXEC)
 
 
-def get_typed_attr(obj, attr, _type, default=None):
+@overload
+def get_typed_attr(obj: Any, attr: str, _type: type[T], default: V) -> T | V:
+    ...
+
+
+@overload
+def get_typed_attr(
+    obj: Any, attr: str, _type: type[T], default: None = None
+) -> T | None:
+    ...
+
+
+def get_typed_attr(
+    obj: Any, attr: str, _type: type[T], default: Optional[V] = None
+) -> T | V | None:
     """
     Typecasts an object's named attribute. If the attribute cannot be
     converted, the default value is returned instead.
@@ -573,13 +630,13 @@ def get_typed_attr(obj, attr, _type, default=None):
     default: value, optional
     """
     try:
-        val = _type(getattr(obj, attr, default))
+        val = _type(getattr(obj, attr, default))  # type: ignore[call-arg]
     except (TypeError, ValueError):
         return default
     return val
 
 
-def get_datetime(date, time, *, microseconds=True):
+def get_datetime(date: str, time: str, *, microseconds: bool = True) -> str:
     """
     Combine date and time from dicom to isoformat.
 
@@ -599,17 +656,38 @@ def get_datetime(date, time, *, microseconds=True):
         if fraction was provided in 'time', and 'microseconds' was
         True.
     """
-    if '.' not in time:
+    if "." not in time:
         # add dummy microseconds if not available for strptime to parse
-        time += '.000000'
-    td = time + ':' + date
-    datetime_str = datetime.strptime(td, '%H%M%S.%f:%Y%m%d').isoformat()
+        time += ".000000"
+    td = time + ":" + date
+    datetime_str = datetime.strptime(td, "%H%M%S.%f:%Y%m%d").isoformat()
     if not microseconds:
-        datetime_str = datetime_str.split('.', 1)[0]
+        datetime_str = datetime_str.split(".", 1)[0]
     return datetime_str
 
 
-def remove_suffix(s, suf):
+def strptime_micr(date_string: str, fmt: str) -> datetime:
+    r"""
+    Decorate strptime while supporting optional [.%f] in the format at the end
+
+    Parameters
+    ----------
+    date_string: str
+      Date string to parse
+    fmt: str
+      Format string. If it ends with [.%f], we keep it if date_string ends with
+      '.\d+' regex and not if it does not.
+    """
+
+    optional_micr = "[.%f]"
+    if fmt.endswith(optional_micr):
+        fmt = fmt[: -len(optional_micr)]
+        if re.search(r"\.\d+$", date_string):
+            fmt += ".%f"
+    return datetime.strptime(date_string, fmt)
+
+
+def remove_suffix(s: str, suf: str) -> str:
     """
     Remove suffix from the end of the string
 
@@ -624,11 +702,11 @@ def remove_suffix(s, suf):
         string with "suf" removed from the end (if present)
     """
     if suf and s.endswith(suf):
-        return s[:-len(suf)]
+        return s[: -len(suf)]
     return s
 
 
-def remove_prefix(s, pre):
+def remove_prefix(s: str, pre: str) -> str:
     """
     Remove prefix from the beginning of the string
 
@@ -643,5 +721,5 @@ def remove_prefix(s, pre):
         string with "pre" removed from the beginning (if present)
     """
     if pre and s.startswith(pre):
-        return s[len(pre):]
+        return s[len(pre) :]
     return s
